//      SYMLOAD
//  Copyright 2003 by
//   Dainis Jonitis


#include <stdio.h>



typedef unsigned short USHORT;
typedef unsigned long  ULONG;
typedef unsigned char  UCHAR;


#define SF_32BIT			0x01		// 32-bit symbols
#define SF_BYNAME			0x02		// symbols sorted by value AND by name


#pragma pack (1)
struct MapHeader
{
  USHORT NextMapPara;         // 00 - paragraph of next map (0 for last)
  UCHAR  SymbolFlags;         // 02 - constant symbol type
  UCHAR  _u3;                 // 03 - padding
  USHORT EntryPointSeg;       // 04 - segment of entry point
  USHORT NumOfConstants;      // 06 - number of constants
  USHORT ConstantOffs;        // 08 - offset of constant data (can be truncated)
  USHORT NumOfSegments;       // 10 - number of segments
  USHORT SegmentPara;         // 12 - paragraph of segment chain
  UCHAR  MaxSymbolNameLen;    // 14 - maximum symbol name length
  UCHAR  NameLen;             // 15 - symbol table name length
  UCHAR  Name [1];            // 16 - symbol table name
};

struct LastMapHeader
{
  USHORT NextMapPara;         // 00 - always 0
  UCHAR  Release;             // 02 - MAPSYM release
  UCHAR  Version;             // 03 - MAPSYM version
};

// Version/release info of some SYM generators
//  5.1 - IBM Operating System/2 Symbol File Generator v4.0
//  6.0 - Microsoft Symbol File Generator v6.0
//  4.0 - Nu-Mega MAP File to SYM Converter v2.9


struct SegmentHeader
{
  USHORT NextSegPara;         // 00 - paragraph of next segment (0 for last)
  USHORT NumOfSymbols;        // 02 - number of symbols
  USHORT SymbolOffsets;       // 04 - offset to symbol offset tables (can be truncated)
  USHORT Segment;             // 06 - segment number
  USHORT Instance0;           // 08 - instance 0 physical address
  USHORT Instance1;           // 10 - instance 1 physical address
  USHORT Instance2;           // 12 - instance 2 physical address
  UCHAR  SymbolFlags;         // 14 - segment symbol type
  UCHAR  _u15;                // 15 - padding
  USHORT LineInfoPara;        // 16 - paragraph of line info
  UCHAR  IsSegmentLoaded;     // 18 - is segment loaded
	UCHAR  CurrentInstance;     // 19 - current instance
  UCHAR  NameLen;             // 20 - segment name length
  UCHAR  Name [1];            // 21 - segment name
};

struct SymbolHeader16
{
  USHORT Offset;
  UCHAR  NameLen;
  UCHAR  Name [1];
};

struct SymbolHeader32
{
  ULONG  Offset;
  UCHAR  NameLen;
  UCHAR  Name [1];
};

// Normally the SymbolOffset points to the USHORT table that contains symbol headers offsets.
// This table is sorted by symbol address. If SF_BYNAME flag is set then other table follows
// that is sorted by symbol name.
// Unfortunately these tables can not be used safely because if there are more than 64K of
// symbol headers then all these offsets are truncated. The same applies to the constant table
// in the map header.

struct SourceHeader
{
  USHORT NextSourcePara;      // 00 - paragraph of next source (0 for last)
  USHORT SegmentOffset;       // 02
  USHORT LineNumberOffset;    // 04
  USHORT FileHandle;          // 06 - file handle
  USHORT NumOfLines;          // 08 - number of line number entries
  UCHAR  NameLen;             // 10 - source file name length
  UCHAR  Name;                // 11 - source file name
};

struct LineHeaderV50_16
{
  USHORT Offset;              // 00 - offset from beginning of segment
  ULONG  FileOffset;          // 02 - offset from start of file (0xFFFFFFFF - not used)
};

struct LineHeaderV50_32       // Not tested !!!
{
  ULONG  Offset;              // 00 - offset from beginning of segment
  ULONG  FileOffset;          // 04 - offset from start of file (0xFFFFFFFF - not used)
};

struct LineHeaderV60_16
{
  USHORT Offset;              // 00 - offset from beginning of segment
  USHORT LineNumber;          // 02 - line number
};

struct LineHeaderV60_32       // Not tested !!!
{
  ULONG  Offset;              // 00 - offset from beginning of segment
  USHORT LineNumber;          // 04 - line number
};
#pragma pack ()



int FileLen (FILE* pFile);
ULONG GetSymbolDataSize (SegmentHeader* pSegment);
UCHAR* LookupSymbol (SegmentHeader* pSegment, ULONG Offset);



int main (int argc, char* argv[])
{
  if (argc != 2)
  {
    fprintf (stderr, "SymLoad file.sym\n");
    return 1;
  }
  
  FILE* pSymFile = fopen (argv [1], "rb");
  if (pSymFile == NULL)
  {
    fprintf (stderr, "Can not open '%s'\n", argv [1]);
    return 2;
  }


  int SymFileLen = FileLen (pSymFile);
  UCHAR* pBuf = new UCHAR [SymFileLen];

  if (pBuf == NULL)
  {
    fprintf (stderr, "Can not allocate memory\n");
    fclose (pSymFile);
    return 3;
  }

  if (fread (pBuf, 1, SymFileLen, pSymFile) == 0)
  {
    fprintf (stderr, "Can not read SYM file\n");
    fclose (pSymFile);
    return 4;
  }  

  fclose (pSymFile);



  MapHeader* pMap = reinterpret_cast <MapHeader*> (pBuf);
  USHORT SymbolTables = 0;


  while (pMap->NextMapPara)
  {
    SymbolTables++;
	  pMap = reinterpret_cast <MapHeader*> (pBuf + pMap->NextMapPara * 16);
  }

  LastMapHeader* pLastMap = reinterpret_cast <LastMapHeader*> (pMap);
  printf ("MAPSYM version: %d.%d\n", pLastMap->Version, pLastMap->Release);
  printf ("Symbol tables:  %d\n", SymbolTables);



  pMap = reinterpret_cast <MapHeader*> (pBuf);

	while (pMap->NextMapPara)
	{
		printf ("\nSymbol map: %.*s\n", pMap->NameLen, &pMap->Name);
		printf (" +00: %04Xh - Next map paragraph\n", pMap->NextMapPara);
		printf (" +02: %02Xh   - Global symbol flags: %s, %s\n",
						pMap->SymbolFlags,
					  (pMap->SymbolFlags & SF_32BIT) ? "32-bit" : "16-bit",
					  (pMap->SymbolFlags & SF_BYNAME) ? "by name and value" : "by value only");
  	printf (" +03: %02Xh   - pad\n", pMap->_u3);
		printf (" +04: %04Xh - Segment of entry point\n", pMap->EntryPointSeg);
		printf (" +06: %-5d - Number of constants\n", pMap->NumOfConstants);
		printf (" +08: %04Xh - Offset of constant data\n", pMap->ConstantOffs, pMap->ConstantOffs);
		printf (" +10: %-5d - Number of segments\n", pMap->NumOfSegments);
		printf (" +12: %04Xh - Paragraph of segment chain\n", pMap->SegmentPara);
		printf (" +14: %-5d - Max symbol name length\n", pMap->MaxSymbolNameLen);
		printf (" +15: %-5d - Module name len\n", pMap->NameLen);



		if (pMap->NumOfConstants)
		{
      int MapHeadSize = sizeof (MapHeader) - sizeof (pMap->Name) + pMap->NameLen;
      UCHAR* pSymbol = reinterpret_cast <UCHAR*> (pMap) + MapHeadSize;


			printf ("\nConstants by Value\n");

			for (USHORT cnt = 0 ; cnt < pMap->NumOfConstants ; cnt++)
			{
        ULONG SymEntrySize;
        
        if (pMap->SymbolFlags & SF_32BIT)
				{
          SymbolHeader32* pSymbol32 = reinterpret_cast <SymbolHeader32*> (pSymbol);
          SymEntrySize = 5 + reinterpret_cast <SymbolHeader32*> (pSymbol)->NameLen;

					printf (" %08X   %.*s\n", 
									pSymbol32->Offset,
									pSymbol32->NameLen, &pSymbol32->Name);
				} else
				{
					SymbolHeader16* pSymbol16 = reinterpret_cast <SymbolHeader16*> (pSymbol);
          SymEntrySize = 3 + reinterpret_cast <SymbolHeader16*> (pSymbol)->NameLen;

					printf (" %04X   %.*s\n", 
									pSymbol16->Offset,
									pSymbol16->NameLen, &pSymbol16->Name);
				}

        pSymbol += SymEntrySize;
      }
		}



		SegmentHeader* pSegment = (pMap->SegmentPara) ? reinterpret_cast <SegmentHeader*> (pBuf + pMap->SegmentPara * 16) : NULL;
    USHORT NumOfSegments = 0;
    USHORT PrevSegParagraph = 0;

		while (pSegment)
		{
			printf ("\nSegment %d - %.*s\n", pSegment->Segment, pSegment->NameLen, &pSegment->Name);
			printf (" +00: %04Xh - Next segment paragraph\n", pSegment->NextSegPara);
			printf (" +02: %-5d - Number of symbols\n", pSegment->NumOfSymbols);
			printf (" +04: %04Xh - Offset of symbol offsets\n", pSegment->SymbolOffsets, pSegment->SymbolOffsets);
			printf (" +06: %04Xh - Segment number\n", pSegment->Segment);
			printf (" +08: %04Xh - Instance 0 physical address\n", pSegment->Instance0);
			printf (" +10: %04Xh - Instance 1 physical address\n", pSegment->Instance1);
			printf (" +12: %04Xh - Instance 2 physical address\n", pSegment->Instance2);
			printf (" +14: %02Xh   - Segment symbol type: %s, %s\n",
				  		pSegment->SymbolFlags,
					   	(pSegment->SymbolFlags & SF_32BIT) ? "32-bit" : "16-bit",
						  (pSegment->SymbolFlags & SF_BYNAME) ? "by name and value" : "by value only");
			printf (" +15: %02Xh   - pad\n", pSegment->_u15);
			printf (" +16: %04Xh - Line number paragraph\n", pSegment->LineInfoPara);
			printf (" +18: %02Xh   - Is segment loaded\n", pSegment->IsSegmentLoaded);
			printf (" +19: %02Xh   - Current instance\n", pSegment->CurrentInstance);
			printf (" +20: %-5d - Segment name len\n", pSegment->NameLen);



			if (pSegment->NumOfSymbols)
			{
        ULONG DataSize = GetSymbolDataSize (pSegment);
        if (DataSize > 0xFFFF)
          fprintf (stderr, "Symbol data exceeds 64K (%d bytes)\n", DataSize);
        

        int SegHeadSize = sizeof (SegmentHeader) - sizeof (pSegment->Name) + pSegment->NameLen;
        UCHAR* pSymbol = reinterpret_cast <UCHAR*> (pSegment) + SegHeadSize;


				printf ("\nPublics by Value\n");

				for (USHORT cnt = 0 ; cnt < pSegment->NumOfSymbols ; cnt++)
				{
          ULONG SymEntrySize;
          
          if (pSegment->SymbolFlags & SF_32BIT)
					{
            SymbolHeader32* pSymbol32 = reinterpret_cast <SymbolHeader32*> (pSymbol);
            SymEntrySize = 5 + reinterpret_cast <SymbolHeader32*> (pSymbol)->NameLen;

						printf (" %04X:%08X   %.*s\n", 
										pSegment->Segment, pSymbol32->Offset,
										pSymbol32->NameLen, &pSymbol32->Name);
					} else
					{
						SymbolHeader16* pSymbol16 = reinterpret_cast <SymbolHeader16*> (pSymbol);
            SymEntrySize = 3 + reinterpret_cast <SymbolHeader16*> (pSymbol)->NameLen;

						printf (" %04X:%04X   %.*s\n", 
										pSegment->Segment, pSymbol16->Offset,
										pSymbol16->NameLen, &pSymbol16->Name);
					}

          pSymbol += SymEntrySize;
        }
			}
	

      
      if (pSegment->LineInfoPara)
      {
        SourceHeader* pSource = reinterpret_cast <SourceHeader*> (pBuf + pSegment->LineInfoPara * 16);
      

        while (pSource)
        {
				  printf ("\nSource file: %.*s\n", pSource->NameLen, &pSource->Name);
          printf (" +00: %04Xh - Next source paragraph\n", pSource->NextSourcePara);
          printf (" +02: %04Xh - Pointer to segment\n", pSource->SegmentOffset);
          printf (" +04: %04Xh - Pointer to line numbers\n", pSource->LineNumberOffset);
          printf (" +06: %04Xh - File handle\n", pSource->FileHandle);
          printf (" +08: %-5d - Line number entries\n", pSource->NumOfLines);
          printf (" +10: %-5d - Source file name len\n", pSource->NameLen);


          printf ("\nLine numbers for segment %.*s\n", pSegment->NameLen, pSegment->Name);
          
          if (pLastMap->Version == 5)
          {
            if (pSegment->SymbolFlags & SF_32BIT)
            {
              LineHeaderV50_32* pLineHead = reinterpret_cast <LineHeaderV50_32*> (reinterpret_cast <UCHAR*> (pSource) + pSource->LineNumberOffset);
              ULONG PrevOffset = ~pLineHead->Offset;

              for (USHORT line = 1 ; line <= pSource->NumOfLines ; line++)
              {
                if (pLineHead->Offset != PrevOffset)
                {
                  PrevOffset = pLineHead->Offset;
                
                  printf (" line %-5d   %04X:%08X", line, pSegment->Segment, pLineHead->Offset);

                  UCHAR* pSymbolName = LookupSymbol (pSegment, pLineHead->Offset);

                  if (pSymbolName)
                    printf ("  (%.*s)\n", pSymbolName [0], &pSymbolName [1]);
                  else
                    printf ("\n");
                }

                if (pLineHead->FileOffset != 0xFFFFFFFF)
                  printf ("  line %-5d   file offset: %08Xh\n", line, pLineHead->FileOffset);


                pLineHead++;
              }
            } else
            {
              LineHeaderV50_16* pLineHead = reinterpret_cast <LineHeaderV50_16*> (reinterpret_cast <UCHAR*> (pSource) + pSource->LineNumberOffset);
              USHORT PrevOffset = ~pLineHead->Offset;

              for (USHORT line = 1 ; line <= pSource->NumOfLines ; line++)
              {
                if (pLineHead->Offset != PrevOffset)
                {
                  PrevOffset = pLineHead->Offset;
                
                  printf (" line %-5d   %04X:%04X", line, pSegment->Segment, pLineHead->Offset);

                  UCHAR* pSymbolName = LookupSymbol (pSegment, pLineHead->Offset);
                  if (pSymbolName)
                    printf ("  (%.*s)\n", pSymbolName [0], &pSymbolName [1]);
                  else
                    printf ("\n");
                }

                if (pLineHead->FileOffset != 0xFFFFFFFF)
                  printf ("  line %-5d   file offset: %08Xh\n", line, pLineHead->FileOffset);


                pLineHead++;
              }
            }
          } else 
          if (pLastMap->Version == 6)
          {
            if (pSegment->SymbolFlags & SF_32BIT)
            {
              LineHeaderV60_32* pLineHead = reinterpret_cast <LineHeaderV60_32*> (reinterpret_cast <UCHAR*> (pSource) + pSource->LineNumberOffset);

              for (USHORT l = 0 ; l < pSource->NumOfLines ; l++)
              {
                printf (" line %-5d   %04X:%08X", pLineHead->LineNumber, pSegment->Segment, pLineHead->Offset);

                UCHAR* pSymbolName = LookupSymbol (pSegment, pLineHead->Offset);
                if (pSymbolName)
                  printf ("  (%.*s)\n", pSymbolName [0], &pSymbolName [1]);
                else
                  printf ("\n");


                pLineHead++;
              }
            } else
            {
              LineHeaderV60_16* pLineHead = reinterpret_cast <LineHeaderV60_16*> (reinterpret_cast <UCHAR*> (pSource) + pSource->LineNumberOffset);

              for (USHORT l = 0 ; l < pSource->NumOfLines ; l++)
              {
                printf (" line %-5d   %04X:%04X", pLineHead->LineNumber, pSegment->Segment, pLineHead->Offset);

                UCHAR* pSymbolName = LookupSymbol (pSegment, pLineHead->Offset);
                if (pSymbolName)
                  printf ("  (%.*s)\n", pSymbolName [0], &pSymbolName [1]);
                else
                  printf ("\n");


                pLineHead++;
              }
            }
          }
          
          pSource = (pSource->NextSourcePara) ? reinterpret_cast <SourceHeader*> (pBuf + pSource->NextSourcePara * 16) : NULL;
        } // next source file
      }
      
      
      
      NumOfSegments++;


      // Nu-Mega's NMSYM.EXE chains segments in circular list and does not have terminating 0 in the last segment.
      // To avoid infinite loop make sure that segment offsets are sequentially increasing.

      if (pSegment->NextSegPara > PrevSegParagraph)
      {
        PrevSegParagraph = pSegment->NextSegPara;
        
        pSegment = (pSegment->NextSegPara) ? reinterpret_cast <SegmentHeader*> (pBuf + pSegment->NextSegPara * 16) : NULL;
      } else
        pSegment = NULL;
		}	// next segment


		if (NumOfSegments != pMap->NumOfSegments)
			fprintf (stderr, "WARNING: Number of real and specified segments does not mach (%d/%d)\n", NumOfSegments, pMap->NumOfSegments);


	  pMap = (pMap->NextMapPara) ? reinterpret_cast <MapHeader*> (pBuf + pMap->NextMapPara * 16) : NULL;
	}	// next symbol map


  delete [] pBuf;
  
  return 0;
}



ULONG GetSymbolDataSize (SegmentHeader* pSegment)
{
  int SegHeadSize = sizeof (SegmentHeader) - sizeof (pSegment->Name) + pSegment->NameLen;
  UCHAR* pSymbol = reinterpret_cast <UCHAR*> (pSegment) + SegHeadSize;
  ULONG DataSize = 0;

  for (USHORT cnt = 0 ; cnt < pSegment->NumOfSymbols ; cnt++)
  {
    ULONG SymEntrySize = (pSegment->SymbolFlags & SF_32BIT) ?
      5 + reinterpret_cast <SymbolHeader32*> (pSymbol)->NameLen :
      3 + reinterpret_cast <SymbolHeader16*> (pSymbol)->NameLen;

    pSymbol += SymEntrySize;
    DataSize += SymEntrySize;
  }

  return DataSize;
}


UCHAR* LookupSymbol (SegmentHeader* pSegment, ULONG Offset)
{
  int SegHeadSize = sizeof (SegmentHeader) - sizeof (pSegment->Name) + pSegment->NameLen;
  UCHAR* pSymbol = reinterpret_cast <UCHAR*> (pSegment) + SegHeadSize;

  for (USHORT cnt = 0 ; cnt < pSegment->NumOfSymbols ; cnt++)
  {
    ULONG SymEntrySize;
    ULONG SymOffset;
    UCHAR* pSymNameLen;

    if (pSegment->SymbolFlags & SF_32BIT)
    {
      SymbolHeader32* pSymbol32 = reinterpret_cast <SymbolHeader32*> (pSymbol);

      SymEntrySize = 5 + pSymbol32->NameLen;
      SymOffset = pSymbol32->Offset;
      pSymNameLen = &pSymbol32->NameLen;
    } else
    {
      SymbolHeader16* pSymbol16 = reinterpret_cast <SymbolHeader16*> (pSymbol);

      SymEntrySize = 3 + pSymbol16->NameLen;
      SymOffset = pSymbol16->Offset;
      pSymNameLen = &pSymbol16->NameLen;
    }

    if (SymOffset == Offset)
      return pSymNameLen;

      
    pSymbol += SymEntrySize;
  }

  return NULL;
}

int FileLen (FILE* pFile)
{
  fseek (pFile, 0, SEEK_END);
  int EndPos = ftell (pFile);
  fseek (pFile, 0, SEEK_SET);

  return EndPos;
}


